/* arch/arm/mach-s3c2410/mach-tomtomgo.c
 *
 * Machine support for the TomTom GO, an all-in-one car navigation product.
 * See <http://www.tomtomgo.com/>.
 *
 * Copyright (C) 2004,2005 TomTom BV <http://www.tomtom.com/>
 * Authors: Dimitry Andric <dimitry.andric@tomtom.com>
 *          Jeroen Taverne <jeroen.taverne@tomtom.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include <linux/kernel.h>
#include <linux/types.h>
#include <linux/interrupt.h>
#include <linux/list.h>
#include <linux/timer.h>
#include <linux/init.h>
#include <linux/device.h>

#include <asm/mach/arch.h>
#include <asm/mach/map.h>
#include <asm/mach/irq.h>

#include <asm/hardware.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/mach-types.h>

#include <asm/arch/regs-serial.h>
#include <asm/arch/regs-gpio.h>
#include <asm/arch/regs-irq.h>
#include <asm/arch/regs-clock.h>
#include <asm/arch/tomtomgo-irq.h>
#include <asm/arch/tomtomgo-map.h>
#include <barcelona/gopins.h>

#include "s3c2410.h"
#include "s3c2440.h"
#include "clock.h"
#include "devs.h"
#include "tomtomgo-devs.h"
#include "cpu.h"
#include "pm.h"

static struct map_desc tomtomgo_iodesc[] __initdata = {
	{ (u32)S3C24XX_VA_USBDEV, S3C2410_PA_USBDEV, S3C24XX_SZ_USBDEV, MT_DEVICE },
	{ (u32)S3C24XX_VA_SDI,    S3C2410_PA_SDI,    S3C24XX_SZ_SDI,    MT_DEVICE },
	{ (u32)S3C24XX_VA_IIS,    S3C2410_PA_IIS,    S3C24XX_SZ_IIS,    MT_DEVICE },
	{ (u32)TOMTOMGO_VA_IDE,   TOMTOMGO_PA_IDE,   TOMTOMGO_SZ_IDE,   MT_DEVICE },
};

#define UCON	(S3C2410_UCON_TXILEVEL  | \
		 S3C2410_UCON_RXILEVEL  | \
		 S3C2410_UCON_TXIRQMODE | \
		 S3C2410_UCON_RXIRQMODE | \
		 S3C2410_UCON_RXFIFO_TOI)
#define ULCON	(S3C2410_LCON_CS8 | \
		 S3C2410_LCON_PNONE)
#define UFCON	(S3C2410_UFCON_FIFOMODE | \
		 S3C2410_UFCON_TXTRIG0  | \
		 S3C2410_UFCON_RXTRIG12)

static struct s3c2410_uartcfg tomtomgo_uartcfgs[] = {
	[0] = {
		.hwport	     = 0,
		.flags	     = 0,
		.ucon	     = UCON,
		.ulcon	     = ULCON,
		.ufcon	     = UFCON,
	},
	[1] = {
		.hwport	     = 1,
		.flags	     = 0,
		.ucon	     = UCON,
		.ulcon	     = ULCON,
		.ufcon	     = UFCON,
	},
	[2] = {
		.hwport	     = 2,
		.flags	     = 0,
		.ucon	     = UCON,
		.ulcon	     = ULCON,
		.ufcon	     = UFCON,
	}
};

/* Standard TomTom GO devices */

static struct platform_device *tomtomgo_devices[] __initdata = {
	&s3c_device_usb,
	&s3c_device_lcd,
	&s3c_device_wdt,
	&s3c_device_i2c,
	&s3c_device_iis,
	&s3c_device_rtc,
	&s3c_device_sdi,
	&s3c_device_usbgadget,
//#if defined CONFIG_BARCELONA_IDE || defined CONFIG_BARCELONA_IDE_MODULE
	&tomtomgo_device_ide,
//#endif /* CONFIG_BARCELONA_IDE || CONFIG_BARCELONA_IDE_MODULE */
#if defined CONFIG_BARCELONA_HW || defined CONFIG_BARCELONA_HW_MODULE
	&tomtomgo_device_hw,
#endif /* CONFIG_BARCELONA_HW || CONFIG_BARCELONA_HW_MODULE */
#if defined CONFIG_BARCELONA_ADC || defined CONFIG_BARCELONA_ADC_MODULE
	&tomtomgo_device_adc,
#endif /* CONFIG_BARCELONA_ADC || CONFIG_BARCELONA_ADC_MODULE */
#if defined CONFIG_BARCELONA_ACC || defined CONFIG_BARCELONA_ACC_MODULE
	&tomtomgo_device_acc,
#endif /* CONFIG_BARCELONA_ACC || CONFIG_BARCELONA_ACC_MODULE */
#if defined CONFIG_BARCELONA_BAT || defined CONFIG_BARCELONA_BAT_MODULE
	&tomtomgo_device_bat,
#endif /* CONFIG_BARCELONA_BAT || CONFIG_BARCELONA_BAT_MODULE */
#if defined CONFIG_BARCELONA_GPIO || defined CONFIG_BARCELONA_GPIO_MODULE
	&tomtomgo_device_gpio,
#endif /* CONFIG_BARCELONA_GPIO || CONFIG_BARCELONA_GPIO_MODULE */
#if defined CONFIG_BARCELONA_GPS || defined CONFIG_BARCELONA_GPS_MODULE
	&tomtomgo_device_gps,
#endif /* CONFIG_BARCELONA_GPS || CONFIG_BARCELONA_GPS_MODULE */
#if defined CONFIG_BARCELONA_PWM || defined CONFIG_BARCELONA_PWM_MODULE
	&tomtomgo_device_pwm,
#endif /* CONFIG_BARCELONA_PWM || CONFIG_BARCELONA_PWM_MODULE */
#if defined CONFIG_BARCELONA_RC || defined CONFIG_BARCELONA_RC_MODULE
	&tomtomgo_device_rc,
#endif /* CONFIG_BARCELONA_RC || CONFIG_BARCELONA_RC_MODULE */
#if defined CONFIG_BARCELONA_SD || defined CONFIG_BARCELONA_SD_MODULE
	&tomtomgo_device_sd,
#endif /* CONFIG_BARCELONA_SD || CONFIG_BARCELONA_SD_MODULE */
#if defined CONFIG_BARCELONA_TS || defined CONFIG_BARCELONA_TS_MODULE
	&tomtomgo_device_ts,
#endif /* CONFIG_BARCELONA_TS || CONFIG_BARCELONA_TS_MODULE */
};

static struct s3c24xx_board tomtomgo_board __initdata = {
	.devices       = tomtomgo_devices,
	.devices_count = ARRAY_SIZE(tomtomgo_devices)
};

#ifdef CONFIG_SMDK2440_BOARD
#define TOMTOMGO_XTAL 16934400
#else
#define TOMTOMGO_XTAL 12000000
#endif

static void __init tomtomgo_map_io(void)
{
	s3c24xx_init_io(tomtomgo_iodesc, ARRAY_SIZE(tomtomgo_iodesc));
	s3c24xx_init_clocks(TOMTOMGO_XTAL);
	s3c24xx_init_uarts(tomtomgo_uartcfgs, ARRAY_SIZE(tomtomgo_uartcfgs));
	s3c24xx_set_board(&tomtomgo_board);
	IO_Init();
}

#ifdef CONFIG_PM
static void tomtomgo_power_off(void)
{
	/* Sleep without simulated wake and without resume to Linux */
	s3c2410_pm_sleep(0,0);
}

static void __init tomtomgo_init_machine(void)
{
	s3c2410_pm_init();
	pm_power_off = tomtomgo_power_off;
}
#endif /* CONFIG_PM */

MACHINE_START(TOMTOMGO, "TomTom GO")
	/* Maintainer: Dimitry Andric <dimitry.andric@tomtom.com> */
	.phys_ram     = S3C2410_SDRAM_PA,
	.phys_io      = S3C2410_PA_UART,
	.io_pg_offst  = (((u32)S3C24XX_VA_UART) >> 18) & 0xfffc,
	.boot_params  = S3C2410_SDRAM_PA + 0x0,
	.map_io       = tomtomgo_map_io,
	.init_irq     = s3c24xx_init_irq,
	.timer        = &s3c24xx_timer,
#ifdef CONFIG_PM
	.init_machine = tomtomgo_init_machine,
#endif /* CONFIG_PM */
MACHINE_END
