/* arch/arm/mach-s3c2410/tomtomgo-ioexp.c
 *
 * Implementation of the IO expander driver.
 *
 * Copyright (C) 2006 TomTom BV <http://www.tomtom.com/>
 * Author: Jeroen Taverne <jeroen.taverne@tomtom.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#ifdef __KERNEL__
#include <linux/kernel.h>
#include <linux/config.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/types.h>
#include <linux/wait.h>
#include <linux/slab.h>
#include <linux/fs.h>
#include <linux/sched.h>
#include <linux/poll.h>
#include <linux/device.h>
#include <linux/interrupt.h>
#include <linux/delay.h>
#include <linux/stat.h>
#include <asm/io.h>
#include <asm/uaccess.h>
#include <asm/irq.h>
#include <asm/hardware.h>
#include <asm/hardware/clock.h>
#include <asm/arch/regs-gpio.h>
#include <barcelona/Barc_Gpio.h>
#include <barcelona/gopins.h>
#include <barcelona/debug.h>
#endif /* __KERNEL__ */

#ifdef __BOOTLOADER__
#include "compiler.h"
#include "Timer.h"
#include "gopins.h"
#define EXPORT_SYMBOL(x)
#endif  /* __BOOTLOADER__ */

#include "tomtomgo-iic.h"

#define IOEXP_ADDRESS 	0x40
#define IOEXP_IODIR 	0x00
#define IOEXP_IPOL  	0x01
#define IOEXP_GPINTEN 	0x02
#define IOEXP_DEFVAL 	0x03
#define IOEXP_INTCON 	0x04
#define IOEXP_IOCON 	0x05
#define IOEXP_GPPU 		0x06
#define IOEXP_INTF 		0x07
#define IOEXP_INTCAP 	0x08
#define IOEXP_GPIO 		0x09
#define IOEXP_OLAT 		0x0a

#define IOEXP_INPUT_MASK ~(1 << 2)
#define IOEXP_PULLUP_MASK ~((1 << 5) | (1 << 2))

static unsigned int result = 0xff;
static unsigned int outputValue = 0;

static void ioexp_tasklet(unsigned long data);
#ifdef __KERNEL__
DECLARE_TASKLET(ioexp_tasklet_handle, ioexp_tasklet, 0L);
#endif

#ifdef __KERNEL__
unsigned int IOEXP_InputOutput(unsigned reqOutputValue)
{
	unsigned int returnResult;

	tasklet_disable(&ioexp_tasklet_handle);
	outputValue = reqOutputValue;
	tasklet_schedule(&ioexp_tasklet_handle);
	returnResult = result;
	tasklet_enable(&ioexp_tasklet_handle);
	return returnResult;
}

unsigned int IOEXP_InputOutputDirect(unsigned reqOutputValue)
{
	unsigned int returnResult;

	outputValue = reqOutputValue;
	ioexp_tasklet(0);
	returnResult = result;
	return returnResult;
}
#endif

#ifdef __BOOTLOADER__
unsigned int IOEXP_InputOutput(unsigned reqOutputValue)
{
	unsigned int returnResult;

	outputValue = reqOutputValue;
	ioexp_tasklet(0);
	returnResult = result;
	return returnResult;
}
#endif

static void ioexp_tasklet(unsigned long data)
{
	static int expanderDetected = 0;
	static int dockint;

	IO_Activate(DOCK_PWREN);
	
	IO_SetInput(SW_SDA);
	
	dockint = IO_GetInput(DOCK_INT);

	expanderDetected = 0;
	if ((dockint) && (IO_GetInput(SW_SDA)))
	{
		IOIIC_Start();
		if (IOIIC_Transmit(IOEXP_ADDRESS)) 
		{
			expanderDetected = 1; // any of &= will reset expanderDetected
			expanderDetected &= IOIIC_Transmit(IOEXP_IODIR);		// Start with first register
			expanderDetected &= IOIIC_Transmit(IOEXP_INPUT_MASK);	// IODIR	Set input pins
			expanderDetected &= IOIIC_Transmit(0x00);				// IPOL		No pin inversion
			expanderDetected &= IOIIC_Transmit(0x00);				// GPINTEN	Set interrupt on inputs
			expanderDetected &= IOIIC_Transmit(0x00);				// DEFVAL	Reset default value
			expanderDetected &= IOIIC_Transmit(0x00);				// INTCON	Compare with last pin value
			expanderDetected &= IOIIC_Transmit(0x00);				// IOCON	Active low interrupt (DOCK_INT will be low when no change is detected)
			expanderDetected &= IOIIC_Transmit(IOEXP_PULLUP_MASK);	// GPPU		Enable pullups for input pins
			expanderDetected &= IOIIC_Transmit(0x00);				// INTF		Read only register
			expanderDetected &= IOIIC_Transmit(0x00);				// INTCAP	Read only register
			expanderDetected &= IOIIC_Transmit(outputValue);		// GPIO		Send output data to GPIO register
			IOIIC_Stop();
			IOIIC_Start();
			expanderDetected &= IOIIC_Transmit(IOEXP_ADDRESS);		// Send chip address
			expanderDetected &= IOIIC_Transmit(IOEXP_GPIO);			// GPIO	Select GPIO register
			IOIIC_Stop();
			IOIIC_Start();
			expanderDetected &= IOIIC_Transmit(IOEXP_ADDRESS+1);	// Send chip address
			
			result = IOIIC_Receive(0);			// Receive GPIO data
			result |= (1 << 9);					// Add virtual DOCK_SENSE pin signal
		}
		IOIIC_Stop();
	}

	dockint = IO_GetInput(DOCK_INT);

	// Check if IO expander is still connected
	if (!dockint)
	{
		expanderDetected = 0;
		result = 0xff;
	}

	// Check USB and desk dock if IO expander isn't connected
	// See IOEXP_InitVirtualPins above
	if (!expanderDetected)
	{
		// mark all IIC pins invalid
		result = 0xff;

		// SDA is pulled down permanently in DESKDOCK
		if (!IO_GetInput(SW_SDA))
		{
			result |= (1 << 10); // add virtual DOCK_DESK_SENSE
			// dockint == (sometimes)USB_DETECT on deskdock of valencia
			if (dockint)
				result |= (1 << 8); // add virtual USB_HOST_DETECT
		}
	}
}

