#include <asm/hardware.h>
#include <asm/mach-types.h>
#include <asm/hardware/clock.h>
#include <asm/arch-s3c2410/regs-udc.h>
#include <asm/io.h>
#include <linux/kernel.h>
#include <linux/err.h>
#include <linux/delay.h>
#include <linux/device.h>
#include <linux/interrupt.h>

#include <asm/arch/regs-gpio.h>
#include <asm/arch/regs-clock.h>

#include <barcelona/gopins.h>

#include "buspower.h"

#define USB_STATE_UNDEF   		0
#define USB_STATE_HOST   		4
#define USB_STATE_CLA     		5
#define USB_STATE_NOVBUS  		6
#define USB_STATE_CONFIGURED    7
#define USB_STATE_BUSY		    8

static int usb_flag       = 0;
static int usb_hstate     = USB_STATE_UNDEF;

//USB_INT_REG / USB_INT_EN_REG
#define SUSPEND_INT            	          0x01
#define RESUME_INT                        0x02
#define RESET_INT                         0x04

/* faking devices to draw USB host power not implemented (yet) */
#undef SUPPORT_FAKE_DEVICE

int usbResetTick;

#define PFX "buspower: "

static irqreturn_t buspower_irq_handler(int irq, void *dev_id, struct pt_regs *regs);

static struct clk *clk;

void buspower_SetUsbCurrent(int cur);


// static 
int buspower_probe(struct device *_dev)
{
	int ret;

	disable_irq(IRQ_USBD);

	/* irq */
	ret = request_irq(IRQ_USBD, buspower_irq_handler, 0,  "Buspower IRQ driver", _dev);
	if (ret != 0) {
		printk(KERN_INFO PFX "failed to install irq (%d)\n", ret);
		return ret;
	}

	clk = clk_get(NULL, "usb-device");
	if (IS_ERR(clk)) {
		printk(KERN_INFO PFX "cannot get usb-device clock\n");
		ret = -ENOENT;
		goto err;
	}
	clk_use(clk);
	buspower_hw_init();

	usb_flag 	= 0;
	usb_hstate 	= USB_STATE_UNDEF;

	return 0;

err:
	free_irq(IRQ_USBD, _dev);
	return ret;
		
}

// static 
int buspower_remove(struct device *_dev)
{
	printk(KERN_INFO PFX "%s ()\n", __FUNCTION__);
	
	buspower_hw_exit();
	
	printk(KERN_INFO PFX "Freeing IRQ\n");
	free_irq(IRQ_USBD, _dev);

	/* no unregistration of chardev */

	printk(KERN_INFO PFX "Done\n");
	return 0;
}

static irqreturn_t buspower_irq_handler(int irq, void *dev_id, struct pt_regs *regs)
{
	int handled;

	do {
		int saveIdx = __raw_readb(S3C2410_UDC_INDEX_REG);
		int usb_status = __raw_readb(S3C2410_UDC_USB_INT_REG);
		int usbd_status = __raw_readb(S3C2410_UDC_EP_INT_REG);
	
		handled = 0;
				
		/* ReSeT Interrupt Request - USB reset */
		if (usb_status & S3C2410_UDC_USBINT_RESET) {
			/* clear interrupt */
			__raw_writeb(S3C2410_UDC_USBINT_RESET,
				S3C2410_UDC_USB_INT_REG);
			buspower_ResetDetected();
			handled = 1;
		}

		/* RESume Interrupt Request */
		if (usb_status & S3C2410_UDC_USBINT_RESUME) {
			/* clear interrupt */
			__raw_writel(S3C2410_UDC_USBINT_RESUME,
				S3C2410_UDC_USB_INT_REG);
			handled = 1;
		}

		/* SUSpend Interrupt Request */
		if (usb_status & S3C2410_UDC_USBINT_SUSPEND) {
			/* clear interrupt */
			__raw_writel(S3C2410_UDC_USBINT_SUSPEND,
				S3C2410_UDC_USB_INT_REG);
				
			// if (!is_usb_connected())
//				stop_activity(dev, dev->driver);
			handled = 1;
		}

		/* EP interrupts */
	   	if (usbd_status) {
  			int	i;

  			/* control traffic */
  			if (usbd_status & S3C2410_UDC_INT_EP0) {
  				/* Clear the interrupt bit by setting it to 1 */
  				__raw_writel(S3C2410_UDC_INT_EP0,S3C2410_UDC_EP_INT_REG);
  				handled = 1;
  			}

  			/* endpoint data transfers */
  			for (i = 1; i < 5; i++) {
  				u32 tmp = 1 << i;
  				if (usbd_status & tmp) {
  					/* Clear the interrupt bit by setting it to 1 */
  					__raw_writel(tmp, S3C2410_UDC_EP_INT_REG);
  					handled = 1;
  				}
  			}
		}
		/* restore index reg */
		__raw_writeb(saveIdx, S3C2410_UDC_INDEX_REG);
	} while (handled);
	
	return IRQ_HANDLED;	
}

#define S3C2410_CLKSLOW_USB_CLK_DISABLE		(1 << 7)

void buspower_hw_init(void)
{
	int tmp;

	/* mask usbdevice ints for now */
	disable_irq(IRQ_USBD);

	/* disable EP0-4 USBD interrupts for now */
	__raw_writeb(0x00, S3C2410_UDC_USB_INT_EN_REG);
	
	/* Select USBD instead of USBH1, enable USB port 1 */
	tmp = __raw_readb(S3C2410_MISCCR);
	tmp &= ~S3C2410_MISCCR_USBHOST; // (1<<3);
	tmp &= ~S3C2410_MISCCR_USBSUSPND1; // (1<<13);
	__raw_writeb(tmp, S3C2410_MISCCR);
	
	switch (IO_GetCpuType()) {
		case GOCPU_S3C2410:
		case GOCPU_S3C2440:
		case GOCPU_S3C2442:
			/* 2410/2440/2442, not 2412/2413 */
			tmp = __raw_readb(S3C2410_CLKSLOW);
			tmp &= ~S3C2410_CLKSLOW_USB_CLK_DISABLE; /* enable usb device clock */
			__raw_writeb(tmp, S3C2410_CLKSLOW);
			break;
		case GOCPU_UNDEFINED:
		case GOCPU_S3C2412:
		default:
			break;
	}
	
	/* just clear */
	tmp = __raw_readb(S3C2410_UDC_EP_INT_REG);
	tmp = __raw_readb(S3C2410_UDC_USB_INT_REG);

	/* Enable USB interrupts for RESET and SUSPEND/RESUME on USBD */
	tmp  = S3C2410_UDC_USBINT_RESET | S3C2410_UDC_USBINT_SUSPEND;
	__raw_writeb(tmp, S3C2410_UDC_USB_INT_EN_REG);

	/* enable usbd ints on interupt controller */
	enable_irq(IRQ_USBD);
}

void buspower_hw_exit(void)
{
}

/* external or internal (s3c) callbacks */
void buspower_ResetDetected(void)
{
	/* callback from USB ISR suspend/resume/reset signalling */
	usb_flag = 1;
}

void buspower_SetConfiguredCurrent(void)
{
	/* can called from interrupt context */
	/* callback from USBd if we can draw the full 500 mA (current) */
	// usb_flags = USB_STATE_CONFIGURED;
	
	/* no PM calls, let USB_PollCLA and PM_Poll() take care of that */
}

void buspower_SetDevice(void)
{
	unsigned i;

	IO_SetInput(USB_HOST_DETECT);
	
	/* Select USBD instead of USBH1, enable USB port 1 */
	switch (IO_GetCpuType()) {
		case GOCPU_S3C2412:
			i = __raw_readl(S3C2412_MISCCR);
			i = i & ~(1<<3);
			i = i & ~(1<<13);
			__raw_writel(i, S3C2412_MISCCR);
			break;
		case GOCPU_S3C2410:
		case GOCPU_S3C2440:
		case GOCPU_S3C2442:
			i = __raw_readl(S3C2410_MISCCR);
			i = i & ~(1<<3);
			i = i & ~(1<<13);
			__raw_writel(i, S3C2410_MISCCR);
			break;
		case GOCPU_UNDEFINED:
		default:
			break;
	}
	mdelay(10);

	// Enable USB interrupts for RESET and SUSPEND/RESUME on USBD
	__raw_writeb(0x04 | 0x01, S3C2410_UDC_USB_INT_REG);

	/* Signal to host that USBD is here using D+ pullup (full-speed) */
	IO_Activate(USB_PULL_EN);
}

void buspower_UnSetDevice(void)
{
	// Disable USB interrupts for RESET and SUSPEND/RESUME on USBD
	// __raw_writeb(0x00, S3C2410_UDC_USB_INT_REG);

	IO_Deactivate(USB_PULL_EN);
}

/* end s3c specific stuff */
int buspower_PollCLA(void)
{
	switch (usb_hstate) {
		case USB_STATE_UNDEF:
			if (IO_HaveUsbBusPowered()) {
				buspower_SetUsbCurrent(100);
				if (IO_GetInput(USB_HOST_DETECT)) {
					usb_flag   = 0;
					buspower_SetDevice(); /* wave with carrot to host (enable pullup) */
					usbResetTick = 3 * 5; /* 3 secs no USB activity/reset -> assume CLA */
					usb_hstate = USB_STATE_BUSY;
					printk("Starting USB vbus probe (USB charge supported)\n");
				} else {
					printk("No USB power (1) (USB charge supported)\n");
					buspower_SetUsbCurrent(0);
					usb_hstate = USB_STATE_NOVBUS;
				}
			} else {
				buspower_SetUsbCurrent(0);
				if (IO_GetInput(USB_HOST_DETECT)) {
					printk("USB Host detected (simple USB)\n");
					usb_hstate = USB_STATE_HOST;
				} else {
					printk("USB Host disconnected detected (simple USB)\n");
					usb_hstate = USB_STATE_NOVBUS;
				}
			}
			break;
		case USB_STATE_BUSY:
			if (!IO_GetInput(USB_HOST_DETECT)) {
				printk("No USB power (2) (USB charge supported)\n");
				usb_hstate = USB_STATE_NOVBUS;
				buspower_SetUsbCurrent(100);
			}
			/* timeout */
			if (usbResetTick <= 0) {
				buspower_UnSetDevice(); /* disable pullup etc */
				usb_hstate = USB_STATE_CLA; /* timeout */
				printk("USB CLA power detected (USB charge supported)\n");
			} else {
				if (usb_flag) {
					printk("USB Host detected (USB charge supported)\n");

					usb_hstate = USB_STATE_HOST;
					buspower_UnSetDevice();
				}
			}
			break;
		case USB_STATE_HOST:
			/* end of host mode ? */
			if (!IO_GetInput(USB_HOST_DETECT)) {
				printk("Lost USB host power (3)\n");
				usb_hstate = USB_STATE_NOVBUS;
				buspower_SetUsbCurrent(100);
			}
			break;
		case USB_STATE_CLA:
			if (!IO_GetInput(USB_HOST_DETECT)) {
				printk("Lost USB CLA power (4)\n");
				usb_hstate = USB_STATE_NOVBUS;
				buspower_SetUsbCurrent(0);
			}
			break;
		case USB_STATE_NOVBUS:
			if (IO_GetInput(USB_HOST_DETECT)) {
				/* start polling again */
				usb_hstate = USB_STATE_UNDEF;
			}
			break;
	}
	return usb_hstate;
}

int buspower_ModeDetect()
{
	/* Note: can be called from timer interupt context ! */
	if (IO_GetInput(USB_HOST_DETECT)) {
		/* Vbus present */
		if (!IO_HaveUsbBusPowered()) {
			/* no buspower feature, simple VBUS detect case */
			return USB_MODE_DEVICE;
		} else {
			/* check if we receive reset from host */
			switch (usb_hstate) {
				case USB_STATE_UNDEF:
					buspower_SetUsbCurrent(100); // Host detected, no conf yet
					return 0;
				case USB_STATE_HOST:
					/* USB reset has been received -> host has been detected */
					/* but, do no extra power until user chooses to reboot */
#ifdef SUPPORT_FAKE_DEVICE
					/* IRQ handler will present fake device, wait until IRQ produces USB_STATE_CONFIGURED state -> 500 mA */
					return 0;
#else
					/* no device faking */
					return USB_MODE_HOST;
#endif
				case USB_STATE_CONFIGURED:
					buspower_SetUsbCurrent(500); 	// Host configured
					return USB_MODE_DEVICE;
				case USB_STATE_CLA:
					buspower_SetUsbCurrent(2000);
					return USB_MODE_CLA;
				case USB_STATE_NOVBUS:
					buspower_SetUsbCurrent(100);
					return USB_MODE_NONE;
				default:
					buspower_SetUsbCurrent(100);
					return 0;
			}
		}
	} else {
		usb_hstate = USB_STATE_NOVBUS;
		// Vbus power not present
		return USB_MODE_NONE;
	}
}


/* pm.c */

static int old_usb_current = 0;
static int new_usb_current = 0;


void buspower_SetUsbCurrent(int cur)
{
	/* can be called from inthandler. Signals to poll to set/reduce current */
	new_usb_current = cur;
}

/* set for ES3, reset for ES2 samples */
#define HAS_USBPWR_BYPASS

void buspower_ConfigureUsbCurrent(void)
{
	/* LTC3455/edinburgh implementation */
	// does not touch USB_SUSPEND
	if (0) { // new_usb_current <= 100) {
		IO_Deactivate(USB_PWR_BYPASS);
		IO_Deactivate(USB_HP);
		IO_Deactivate(USB_SUSPEND_OUT);
	} else {
		if (new_usb_current <= 500) {
			IO_Activate(USB_HP);
			IO_Deactivate(USB_PWR_BYPASS);
			IO_Deactivate(USB_SUSPEND_OUT);
		} else {
			if (new_usb_current > 500) {
				#ifdef HAS_USBPWR_BYPASS
					/* order might be important */
					IO_Activate(USB_PWR_BYPASS);
					IO_Deactivate(USB_HP);
					IO_Activate(USB_SUSPEND_OUT);
				#else
					IO_Deactivate(USB_PWR_BYPASS);
					IO_Activate(USB_HP);
					IO_Deactivate(USB_SUSPEND_OUT);
				#endif
			}
		}
	}
}

void buspower_poll(void)
{
	buspower_PollCLA();
	
	if (old_usb_current != new_usb_current) {
		buspower_ConfigureUsbCurrent();
		old_usb_current = new_usb_current;
	}
}

int buspower_suspend(struct device *_dev)
{
	// simple case: suspend always 500 mA mode
	buspower_SetUsbCurrent(500); 
	buspower_ConfigureUsbCurrent();
	
	return 0;
}
