/* drivers/barcelona/hw/hwdrv.c
 *
 * Implementation of the hardware detection driver.
 *
 * Copyright (C) 2004,2005 TomTom BV <http://www.tomtom.com/>
 * Author: Dimitry Andric <dimitry.andric@tomtom.com>
 * Author: Jeroen Taverne <jeroen.taverne@tomtom.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

/* Includes */ 
#include <linux/kernel.h>
#include <linux/config.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/types.h>
#include <linux/wait.h>
#include <linux/slab.h>
#include <linux/fs.h>
#include <linux/sched.h>
#include <linux/poll.h>
#include <linux/device.h>
#include <linux/interrupt.h>
#include <linux/delay.h>
#include <linux/proc_fs.h>
#include <linux/stat.h>
#include <asm/io.h>
#include <asm/uaccess.h>
#include <asm/irq.h>
#include <asm/hardware.h>
#include <asm/hardware/clock.h>
#include <barcelona/gopins.h>
#include <barcelona/debug.h>

/* Defines */
#define PFX "hw: "
#define PK_DBG PK_DBG_FUNC

/* Local variables */
static const char hw_proc_root_name[] = "barcelona";
static struct proc_dir_entry *hw_proc_root;

/* Local types */
struct hw_proc_entry {
	const char *name;
	read_proc_t *func;
};

extern unsigned int reset_state;

static int hw_proc_read(char *page, char **start, off_t offset, int count, int *eof, const char *msg)
{
	int avail = strlen(msg) - offset;

	if (count >= avail) {
		count = avail;
		*eof = 1;
	}
	if (count <= 0)
		return 0;

	*start = page;
	memcpy(page, msg + offset, count);

	return count;
}

static int hw_proc_read_int(char *page, char **start, off_t offset, int count, int *eof, int i)
{
	unsigned char msg[16];
	snprintf(msg, sizeof msg, "%d\n", i);
	return hw_proc_read(page, start, offset, count, eof, msg);
}

static int hw_proc_read_str(char *page, char **start, off_t offset, int count, int *eof, const char *s)
{
	unsigned char msg[128];
	snprintf(msg, sizeof msg, "%s\n", s);
	return hw_proc_read(page, start, offset, count, eof, msg);
}

#define PROC_INT(name, func) \
	int hw_proc_##name(char *page, char **start, off_t offset, int count, int *eof, void *data) \
	{ \
		return hw_proc_read_int(page, start, offset, count, eof, IO_##func()); \
	}

#define PROC_INT_VARIABLE(name, var) \
	int hw_proc_##name(char *page, char **start, off_t offset, int count, int *eof, void *data) \
	{ \
		return hw_proc_read_int(page, start, offset, count, eof, var); \
	}

#define PROC_STR(name, func) \
	int hw_proc_##name(char *page, char **start, off_t offset, int count, int *eof, void *data) \
	{ \
		return hw_proc_read_str(page, start, offset, count, eof, IO_##func()); \
	}

PROC_INT(modelid, GetModelId)
PROC_STR(modelname, GetModelName)
PROC_STR(familyname, GetFamilyName)
PROC_STR(usbname, GetUsbName)
PROC_STR(btname, GetBtName)
PROC_INT(cputype, GetCpuType)
PROC_INT(cpufreq, GetCpuFreq)

PROC_INT(bluetooth, HaveBluetooth)
PROC_INT(btchip, GetBluetoothChip)
PROC_INT(btusb, HaveBluetoothUsb)
PROC_INT(btspeed, GetBluetoothSpeed)
PROC_INT(btclock, GetBluetoothClock)
PROC_INT(btclass, GetBluetoothClass)
PROC_INT(handsfree, HaveHandsfree)
PROC_INT(headsetgw, HaveHeadsetGw)
PROC_INT(a2dp, HaveA2DP)
PROC_INT(lowbutton, HaveLoweredButton)
PROC_INT(battery, HaveBattery)
PROC_INT(sdcard, HaveSdCard)
PROC_INT(harddisk, HaveHarddisk)
PROC_INT(tsfets, HaveTsFets)
PROC_INT(remote, HaveRemote)
PROC_INT(docking, HaveDocking)
PROC_INT(tfttype, GetTftType)
PROC_INT(gpstype, GetGpsType)
PROC_INT(gpsephemeris, HaveGpsEphemeris)
PROC_INT(headphoneconnector, HaveHeadphoneConnector)
PROC_INT(codectype, GetCodecType)
PROC_INT(codecmaster, GetCodecMaster)
PROC_INT(acctype, GetAccType)
PROC_INT(usbslavetype, GetUSBSlaveType)
PROC_INT(lightsensor, HaveLightSensor)
PROC_INT(gyro, HaveGyro)
PROC_INT(loquendo, HaveLoquendo )
PROC_INT_VARIABLE(resetstate, reset_state )

#define HW_PROC_ENTRY(x) { .name = #x, .func = hw_proc_##x }

static struct hw_proc_entry hw_proc_entries[] = {
	HW_PROC_ENTRY(modelid),
	HW_PROC_ENTRY(modelname),
	HW_PROC_ENTRY(familyname),
	HW_PROC_ENTRY(usbname),
	HW_PROC_ENTRY(btname),
	HW_PROC_ENTRY(cputype),
	HW_PROC_ENTRY(cpufreq),
	HW_PROC_ENTRY(bluetooth),
	HW_PROC_ENTRY(btchip),
	HW_PROC_ENTRY(btusb),
	HW_PROC_ENTRY(btspeed),
	HW_PROC_ENTRY(btclock),
	HW_PROC_ENTRY(btclass),
	HW_PROC_ENTRY(handsfree),
	HW_PROC_ENTRY(headsetgw),
	HW_PROC_ENTRY(a2dp),
	HW_PROC_ENTRY(lowbutton),
	HW_PROC_ENTRY(battery),
	HW_PROC_ENTRY(sdcard),
	HW_PROC_ENTRY(harddisk),
	HW_PROC_ENTRY(tsfets),
	HW_PROC_ENTRY(remote),
	HW_PROC_ENTRY(docking),
	HW_PROC_ENTRY(tfttype),
	HW_PROC_ENTRY(gpstype),
	HW_PROC_ENTRY(gpsephemeris),
	HW_PROC_ENTRY(headphoneconnector),
	HW_PROC_ENTRY(codectype),
	HW_PROC_ENTRY(codecmaster),
	HW_PROC_ENTRY(acctype),
	HW_PROC_ENTRY(usbslavetype),
	HW_PROC_ENTRY(lightsensor),
	HW_PROC_ENTRY(gyro),
	HW_PROC_ENTRY(loquendo),
	HW_PROC_ENTRY(resetstate),
};

static int hw_proc_init(void)
{
	int i;

	hw_proc_root = proc_mkdir(hw_proc_root_name, NULL);
	if (hw_proc_root == NULL) {
		PK_ERR("Unable to create hw /proc root %s\n", hw_proc_root_name);
		return -1;
	}

	for (i = 0; i < ARRAY_SIZE(hw_proc_entries); ++i) {
		if (create_proc_read_entry(hw_proc_entries[i].name, S_IRUGO, hw_proc_root, hw_proc_entries[i].func, 0) == NULL) {
			PK_ERR("Unable to create hw /proc entry %s\n", hw_proc_entries[i].name);
			return -1;
		}
	}

	return 0;
}

static void hw_proc_exit(void)
{
	int i;

	if (hw_proc_root != NULL) {
		for (i = ARRAY_SIZE(hw_proc_entries) - 1; i >= 0; --i) {
			remove_proc_entry(hw_proc_entries[i].name, hw_proc_root);
		}
		remove_proc_entry(hw_proc_root_name, NULL);
		hw_proc_root = NULL;
	}
}

static int hw_probe(struct device *dev)
{
	int ret;

	PK_DBG("Setting up proc entries\n");
	ret = hw_proc_init();
	if (ret < 0) {
		PK_ERR("Unable to setup proc entries\n");
		return -ENODEV;
	}

	PK_DBG("Done\n");
	return 0;
}

static int hw_remove(struct device *dev)
{
	PK_DBG("Removing proc entries\n");
	hw_proc_exit();

	PK_DBG("Done\n");
	return 0;
}

static void hw_shutdown(struct device *dev)
{
	PK_DBG("Shutting down\n");
}

#ifdef CONFIG_PM

static int hw_suspend(struct device *dev, u32 state, u32 level)
{
	PK_DBG("dev = %p, state = %u, level = %u\n", dev, state, level);
	if (level == SUSPEND_POWER_DOWN) {
		hw_shutdown(dev);
	}
	return 0;
}

static int hw_resume(struct device *dev, u32 level)
{
	PK_DBG("dev = %p, level = %u\n", dev, level);
	return 0;
}

#else /* CONFIG_PM */
#define hw_suspend NULL
#define hw_resume  NULL
#endif /* CONFIG_PM */

static struct device_driver hw_driver = {
	.name		= "tomtomgo-hw",
	.bus		= &platform_bus_type,
	.probe		= hw_probe,
	.remove		= hw_remove,
	.shutdown	= hw_shutdown,
	.suspend	= hw_suspend,
	.resume		= hw_resume,
};

static int __init hw_mod_init(void)
{
	int ret;

	printk(KERN_INFO "TomTom GO Hardware Detection Driver, (C) 2004,2005 TomTom BV\n");
	PK_DBG("Registering driver\n");
	ret = driver_register(&hw_driver);
	if (ret) {
		PK_ERR("Unable to register driver (%d)\n", ret);
		return ret;
	}
	PK_DBG("Done\n");
	return 0;
}

static void __exit hw_mod_exit(void)
{
	PK_DBG("Unregistering driver\n");
	driver_unregister(&hw_driver);
	PK_DBG("Done\n");
}

module_init(hw_mod_init);
module_exit(hw_mod_exit);

MODULE_AUTHOR("Dimitry Andric <dimitry.andric@tomtom.com> and Jeroen Taverne <jeroen.taverne@tomtom.com>");
MODULE_DESCRIPTION("TomTom GO Hardware Detection Driver");
MODULE_LICENSE("GPL");

/* EOF */
