/* drivers/video/tomtom/ltv350qv.c
 *
 * Specific functionality for the Samsung LTV350QV screen. See:
 * http://www.samsung.com/Products/TFTLCD/Small_n_Medium/LTV350QV_F01/LTV350QV-F01.htm
 *
 * Copyright (C) 2005 TomTom BV <http://www.tomtom.com/>
 * Author: Jeroen Taverne <jeroen.taverne@tomtom.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#ifdef __KERNEL__
#include <linux/delay.h>
#include <barcelona/gopins.h>
#endif /* __KERNEL */
#ifdef __BOOTLOADER__
#include "Timer.h"
#include "gopins.h"
#endif /* __BOOTLOADER__ */
#include "lcdregs.h"
#include "ltv350qv.h"

#define MVAL_USED (0)
#define MVAL (0)

#define VBPD_320240     ((7-1)&0xff)
#define VFPD_320240     ((7-1)&0xff)
#define VSPW_320240     ((3-1) &0x3f)
#define HBPD_320240     ((19-1)&0x7f)
#define HFPD_320240     ((17-1)&0xff)
#define HSPW_320240     ((14-1)&0xff)

#define CLKVAL_TFT_320240 (6) // Maximum dotclock for LTV350QV is 30 MHz

#define LCD_XSIZE_TFT_320240 (320)
#define LCD_YSIZE_TFT_320240 (240)
#define HOZVAL_TFT_320240    (LCD_XSIZE_TFT_320240-1)
#define LINEVAL_TFT_320240   (LCD_YSIZE_TFT_320240-1)

static void LTV350QV_WriteByte(unsigned char data)
{
	int bit;

	for (bit=0;bit<8;bit++)
	{
		IO_Deactivate(LCD_SCL);
		if (data & 0x80) IO_Activate(LCD_SDI); else IO_Deactivate(LCD_SDI);
		IO_Activate(LCD_SCL);
		data <<= 1;
	}
}

static void LTV350QV_Tx_Data(unsigned char Index,unsigned short REG_DATA)
{
	unsigned char DeviceID = 0x74;
	unsigned char lcd_Index = (unsigned char)(Index & 0xff);
	unsigned char lcd_data0 = (unsigned char)((REG_DATA >> 8)& 0xff);
	unsigned char lcd_data1 = (unsigned char)( REG_DATA & 0xff);

	IO_Deactivate(LCD_SCL);
	IO_Deactivate(LCD_CS);

	IO_Activate(LCD_CS);
	LTV350QV_WriteByte(DeviceID|0x0);      //Send Device ID code
	LTV350QV_WriteByte(0x00);	     //Write register 8 bit
	LTV350QV_WriteByte(lcd_Index);     //Write register 8 bit
	IO_Deactivate(LCD_CS);
	mdelay(1);

	IO_Activate(LCD_CS);
	LTV350QV_WriteByte(DeviceID|0x2);      //Send Device ID code
	LTV350QV_WriteByte(lcd_data0);     //Write the data first 8 bit
	LTV350QV_WriteByte(lcd_data1);     //Write the data second 8 bit
	IO_Deactivate(LCD_CS);
	mdelay(1);                 //need to delay 50us
}

void LTV350QV_Init(void)
{
	// Start display controller
	rLCDCON1 &= ~1;
	rTCONSEL = 0;
	rLCDCON1=(CLKVAL_TFT_320240<<8)|(MVAL_USED<<7)|(3<<5)|(12<<1)|0;
	rLCDCON2=(VBPD_320240<<24)|(LINEVAL_TFT_320240<<14)|(VFPD_320240<<6)|(VSPW_320240);
	rLCDCON3=(HBPD_320240<<19)|(HOZVAL_TFT_320240<<8)|(HFPD_320240);
	rLCDCON4=(MVAL<<8)|(HSPW_320240);
	rLCDCON5=(1<<11)|(1<<10)|(1<<9)|(1<<8)|(1<<6)|(1<<0);

	IO_Activate(LCD_RESET);
	mdelay(10);
	IO_Deactivate(LCD_RESET);
	mdelay(10);

	LTV350QV_Tx_Data(0x09,0x0000);
	mdelay(10);
	LTV350QV_Tx_Data(0x09,0x4000);
	LTV350QV_Tx_Data(0x0a,0x2000);
	LTV350QV_Tx_Data(0x09,0x4055);
	mdelay(50);
	LTV350QV_Tx_Data(0x01,0x409d);
	LTV350QV_Tx_Data(0x02,0x0204);    
	LTV350QV_Tx_Data(0x03,0x0100);
	LTV350QV_Tx_Data(0x04,0x3000);
	LTV350QV_Tx_Data(0x05,0x4003);
	LTV350QV_Tx_Data(0x06,0x000a);
	LTV350QV_Tx_Data(0x07,0x0021);
	LTV350QV_Tx_Data(0x08,0x0c00);
	LTV350QV_Tx_Data(0x10,0x0103);
	LTV350QV_Tx_Data(0x11,0x0301);
	LTV350QV_Tx_Data(0x12,0x1f0f);
	LTV350QV_Tx_Data(0x13,0x1f0f);
	LTV350QV_Tx_Data(0x14,0x0707);
	LTV350QV_Tx_Data(0x15,0x0307);
	LTV350QV_Tx_Data(0x16,0x0707);
	LTV350QV_Tx_Data(0x17,0x0000);
	LTV350QV_Tx_Data(0x18,0x0004);
	LTV350QV_Tx_Data(0x19,0x0000);
	rLCDCON1 |= 1;
	mdelay(100);
	LTV350QV_Tx_Data(0x09,0x4a55);
	LTV350QV_Tx_Data(0x05,0x5003);
}

void LTV350QV_Off(void)
{
	LTV350QV_Tx_Data(0x09,0x4055);
	LTV350QV_Tx_Data(0x05,0x4003);    
	mdelay(10);
	LTV350QV_Tx_Data(0x0a,0x0000);
	mdelay(10);
	LTV350QV_Tx_Data(0x09,0x4000);
	mdelay(10);

	// Stop display controller
	rLCDCON1 = 0;

	IO_Activate(LCD_RESET);
	IO_Deactivate(LCD_SCL);
	IO_Deactivate(LCD_SDI);
	IO_Activate(LCD_CS);
}

/* EOF */
