/*
 * File   : kern.h
 * Author : Koen Martens <kmartens@sonologic.nl>
 * Desc.  : Defines and data structures for kernel space only.
 *
 */

#include <linux/list.h>

//
// Defines
//

#define NGFFS_SUPER_MAGIC       0x476d4373

#define NGFFS_GARBAGE_NONE       0
#define NGFFS_GARBAGE_EMPTY      1
#define NGFFS_GARBAGE_ACTIVATE   2
#define NGFFS_GARBAGE_INVALIDATE 3
#define NGFFS_GARBAGE_IGNORE     4
#define NGFFS_GARBAGE_COLLECT    5

//
// Inline functions to convert metadata from ngffs to inode vv.
//

inline static void ngffs_copy_from_meta(struct inode *inode,struct ngffs_metadata *meta) {
  inode->i_mode=meta->mode;                                                       inode->i_nlink=meta->link_count;
  inode->i_uid=meta->uid;                                                         inode->i_gid=meta->gid;
  inode->i_mtime.tv_sec=meta->mtime;                                              inode->i_atime.tv_sec=meta->mtime;
  inode->i_ctime.tv_sec=meta->ctime;                                              inode->i_mtime.tv_nsec=0;
  inode->i_atime.tv_nsec=0;                                                       inode->i_ctime.tv_nsec=0;
}

inline static void ngffs_copy_to_meta(struct inode *inode,struct ngffs_metadata *meta) {
  meta->mode=inode->i_mode;
  meta->link_count=inode->i_nlink;
  meta->uid=inode->i_uid;
  meta->gid=inode->i_gid;
  meta->mtime=inode->i_mtime.tv_sec;
  meta->ctime=inode->i_ctime.tv_sec;
}


// not defined elsewhere, is in lib/string.c
//extern void bcopy(const void * srcp, void * destp, size_t count);

#define bcopy(src,dst,cnt) memcpy(dst,src,cnt)


//
// List data structures
//
struct dir_list_entry {
  struct list_head list;
  struct dir_chunk *chunk;
  __u32 ofs;
};

struct ngffs_data_cache {
  __u32 offset;
  __u32 modification;
};

struct file_list_entry {
  struct list_head list;
  struct file_chunk *chunk;
  __u32 ofs;
  struct ngffs_data_cache *cache;
};

struct ngffs_block {
  struct header_chunk   header;
  __u32                 offset;
  __u32                 size;
  __u32                 freespace;
  __u32                 reserved;       // used when locked, to determine if commit_write may
                                        // succeed
  __u32                 garbage;        // used for flags during garbage collection
};

struct ngffs_info {
  struct ngffs_block *blocklist;
  __u32 offset;                 // offset from start of flash where flash fs starts
  __u32 blocks;                 // number of erase blocks used for fs
  __u32 erasesize;              // total size in bytes of erase blocks
  __u32 size;                   // total size in bytes of fs
  __u32 max_entry_id;           // highest ino in fs
  struct mtd_info *mtd;
  struct list_head dirs;
  struct list_head files;
  rwlock_t global_lock;         // locks the entire filesystem
  struct super_block *sb;       // superblock
};

//
// Some convenient shorthands
//

// Extract the ngffs info from the superblock struct
#define NGFFS_INFO(x)   ((struct ngffs_info *)x->s_fs_info)

// Shorthands to some inode data, not used anymore due to consistency probs
//#define INODE_DIR(x)    (((struct dir_list_entry *)(x->u.generic_ip))->chunk)
//#define INODE_FILE(x)   (((struct file_list_entry *)(x->u.generic_ip))->chunk)
//#define INODE_FILE_OFS(x) (((struct file_list_entry *)(x->u.generic_ip))->ofs)
//#define INODE_DIR_OFS(x) (((struct dir_list_entry *)(x->u.generic_ip))->ofs)


